import configparser
from pathlib import Path

from de.p1st.monitor.cfg.singleton import get_cfg
from de.p1st.monitor.logger_ex import LoggerArgEx
from de.p1st.monitor.loggers.cpu import CPULogger1, CPULogger5, CPULogger15
from de.p1st.monitor.loggers.drive import DriveLogger
from de.p1st.monitor.loggers.filesystem import FilesystemLogger
from de.p1st.monitor.loggers.memory import MemoryLogger
from de.p1st.monitor.loggers.network import NetworkLogger
from de.p1st.monitor.loggers.swap import SwapLogger
from de.p1st.monitor.loggers.temp import TempLogger
from de.p1st.monitor.logger import Logger

def get_or_raise(cfg: configparser.SectionProxy, key: str) -> str:
    if key in cfg:
        return cfg[key]
    else:
        raise LoggerArgEx(f'Missing key {key} in section {cfg.name}')

def get_loggers() -> tuple[list[Logger], list[LoggerArgEx]]:
    def temp(cfg: configparser.SectionProxy) -> Logger:
        sensor = get_or_raise(cfg, 'sensor')
        label = get_or_raise(cfg, 'label')
        warn_if_above = float(cfg['warn_if_above']) if 'warn_if_above' in cfg else None
        warn_threshold = int(cfg.get('warn_threshold', '1'))
        warn_data_range = int(cfg.get('warn_data_range', '1'))
        return TempLogger(sensor, label, warn_if_above, warn_threshold, warn_data_range)
    def cpu1(cfg: configparser.SectionProxy) -> Logger:
        warn_if_above = float(cfg['warn_if_above']) if 'warn_if_above' in cfg else None
        warn_threshold = int(cfg.get('warn_threshold', '1'))
        warn_data_range = int(cfg.get('warn_data_range', '1'))
        return CPULogger1(warn_if_above, warn_threshold, warn_data_range)

    def cpu5(cfg: configparser.SectionProxy) -> Logger:
        warn_if_above = float(cfg['warn_if_above']) if 'warn_if_above' in cfg else None
        warn_threshold = int(cfg.get('warn_threshold', '1'))
        warn_data_range = int(cfg.get('warn_data_range', '1'))
        return CPULogger5(warn_if_above, warn_threshold, warn_data_range)

    def cpu15(cfg: configparser.SectionProxy) -> Logger:
        warn_if_above = float(cfg['warn_if_above']) if 'warn_if_above' in cfg else None
        warn_threshold = int(cfg.get('warn_threshold', '1'))
        warn_data_range = int(cfg.get('warn_data_range', '1'))
        return CPULogger15(warn_if_above, warn_threshold, warn_data_range)

    def net(cfg: configparser.SectionProxy) -> Logger:
        network_interface = get_or_raise(cfg, 'network_interface')
        return NetworkLogger(network_interface)

    def filesystem(cfg: configparser.SectionProxy) -> Logger:
        uuid = cfg.get('uuid', None)
        mountpoint = Path(cfg.get('mountpoint')) if 'mountpoint' in cfg else None
        unmounted_ok = bool(cfg.get('unmounted_ok', 'false'))
        warn_if_above = float(cfg.get('warn_if_above', '1.0'))
        warn_threshold = int(cfg.get('warn_threshold', '1'))
        warn_data_range = int(cfg.get('warn_data_range', '1'))
        return FilesystemLogger(uuid, mountpoint, unmounted_ok, warn_if_above, warn_threshold, warn_data_range)

    def drive(cfg: configparser.SectionProxy) -> Logger:
        uuid = cfg.get('uuid', None)
        device = Path(cfg.get('device')) if 'device' in cfg else None
        warn_if_above = int(cfg['warn_if_above']) if 'warn_if_above' in cfg else None
        warn_threshold = int(cfg.get('warn_threshold', '1'))
        warn_data_range = int(cfg.get('warn_data_range', '1'))
        return DriveLogger(uuid, device, warn_if_above, warn_threshold, warn_data_range)

    def memory(cfg: configparser.SectionProxy) -> Logger:
        warn_if_above = float(cfg.get('warn_if_above', '1.0'))
        warn_threshold = int(cfg.get('warn_threshold', '1'))
        warn_data_range = int(cfg.get('warn_data_range', '1'))
        return MemoryLogger(warn_if_above, warn_threshold, warn_data_range)
    def swap(cfg: configparser.SectionProxy) -> Logger:
        warn_if_above = float(cfg.get('warn_if_above', '1.0'))
        warn_threshold = int(cfg.get('warn_threshold', '1'))
        warn_data_range = int(cfg.get('warn_data_range', '1'))
        return SwapLogger(warn_if_above, warn_threshold, warn_data_range)


    mapping = {
        'temp': temp,
        'cpu1': cpu1,
        'cpu5': cpu5,
        'cpu15': cpu15,
        'network': net,
        'filesystem': filesystem,
        'drive': drive,
        'memory': memory,
        'swap': swap,
    }

    loggers = []
    exceptions = []
    cfg: configparser.ConfigParser = get_cfg()
    for section_name in cfg.sections():
        if section_name == 'logging':
            continue
        prefix = section_name.split('.', maxsplit=1)[0]
        try:
            loggers.append(
                mapping[prefix](cfg[section_name])
            )
        except LoggerArgEx as e:
            exceptions.append(e)

    return loggers, exceptions
