from pathlib import Path
from abc import ABC, abstractmethod

from de.p1st.monitor import csv_util
from de.p1st.monitor.cfg.logging_dir import logging_dir
from de.p1st.monitor.string_conversion import to_string, from_string
from de.p1st.monitor.warn import WarnMessage, WarnLevel


# https://www.geeksforgeeks.org/abstract-classes-in-python/
class Logger(ABC):
    def __init__(self,
                 warn_threshold: int = 1,
                 warn_data_range: int = 1,
                 warn_if_above: int | float = None,
                 critical_if_above: int | float = None,
                 ):
        self.data: list[any] | None = None
        # True if the data held by this object is already appended to the logfile.
        self.logged = False

        self.warn_threshold = warn_threshold
        self.warn_data_range = warn_data_range

        # Either both variables are given, or both are None
        if warn_if_above is not None and critical_if_above is not None:
            assert critical_if_above > warn_if_above
        else:
            assert warn_if_above is None and critical_if_above is None
        self.warn_if_above = warn_if_above
        self.critical_if_above = critical_if_above

    def export_data(self) -> Path:
        """
        This method is intended to be overriden in a subclass!

        With most loggers the `get_log_file()` is ready-to-use.
        In this case this method simply returns `get_log_file()`.

        But some loggers require postprocessing of that data before it can be used.
        In this case this method creates a new .csv file and returns it.

        @return: Path to .csv file with ready-to-use data.
        """
        return self.get_log_file()

    def check(self) -> WarnMessage:
        """
        Checks the latest `self.warn_data_range` datasets for problems using `self.check_data()`.

        If at least `self.warn_threshold` problems are found,
        then a WarnMessage with the highest reported WarnLevel is returned.

        If at least one WarnLevel is above NORMAL,
        then a WarnMessage is returned independent of the number of problems.
        """
        datasets = self.get_datasets(self.warn_data_range)
        warnings = [self.check_data(data) for data in datasets]
        warnings = [warning for warning in warnings
                    if not warning.level == WarnLevel.NONE]

        if len(warnings) == 0:
            return WarnMessage(WarnLevel.NONE)
        # max() must not be called with an empty list.
        highest_warn_level = max([warning.level for warning in warnings])

        messages: list[str] = [warning.message for warning in warnings]
        message = f'{len(warnings)} of the last {self.warn_data_range} datasets are above limits:\n\t' \
                  + '\n\t'.join(messages)

        if highest_warn_level > WarnLevel.NORMAL:
            return WarnMessage(highest_warn_level, message)
        if len(warnings) >= self.warn_threshold:
            return WarnMessage(highest_warn_level, message)
        return WarnMessage(WarnLevel.NONE)

    @abstractmethod
    def check_data(self, data: list[any]) -> WarnMessage:
        """
        Check the given data for problems.
        Return a WarnLevel indicating how serious the problems are.

        If there are no problems, return `WarnLevel.NONE`.
        """
        raise ValueError('Subclasses must implement this')

    def get_all_datasets(self) -> list[list[any]]:
        # See also: self.get_datasets()

        if self.get_log_file().exists():
            # We skip the first row as it is the data schema.
            raw = csv_util.read(self.get_log_file())[1:]
            data = [self.get_data_from_row(row) for row in raw]
        else:
            data = []

        if not self.logged and self.has_data():
            data.append(self.get_data())

        return data

    def get_datasets(self, num: int) -> list[list[any]]:
        """
        Returns the last `num` datasets (including the current dataset).
        """
        if not self.logged and self.has_data():
            # We will append the current data manually.
            # Thus, we need to read one less line from the CSV file.
            read_last = num - 1
        else:
            read_last = num

        if self.get_log_file().exists():
            # Read rows from CSV file.
            # We skip the first row as it is the data schema.
            # We keep only the last `read_last` rows.
            raw = csv_util.read_last(self.get_log_file(), read_last, 1)
            # Convert from string to data types defined in the data schema.
            data = [self.get_data_from_row(row) for row in raw]
        else:
            data = []

        if not self.logged and self.has_data():
            # We append the current data.
            # It has not yet been logged and is therefore not included in the CSV file we just read.
            data.append(self.get_data())

        return data

    def log(self, skip_if_no_data: bool = False) -> None:
        """
        Appends the current data (e.g. temperature of a sensor)
        to a logfile.

        :param skip_if_no_data: Can be used to do nothing if no data is available. If one is sure to have called update() previously, this can be set to True.
        :raise Exception: If method is called but no data is available. Please do call update() first to avoid this!
        """
        if self.logged:
            return
        if skip_if_no_data and not self.has_data():
            return

        csv_util.write(file=self.get_log_file(), rows=[self.get_data_as_row()], header=self.data_schema())
        self.logged = True

    def update(self):
        self.set_data(self.read_data())
        self.logged = False

    @abstractmethod
    def read_data(self) -> list[any] | None:
        """
        Collects current data (e.g. temperature of a sensor).

        Might return None if sensor is detached / not available.

        :raise LoggerReadEx:
        """
        raise ValueError('Subclasses must implement this')

    @abstractmethod
    def data_schema(self) -> list[str]:
        """
        Describes the type and meaning of the elements in self.values().

        Returns a list with elements f'{data-type}#{column-description}'.

        Example:
            ['datetime#Date', 'float#Disk usage']
        """
        raise ValueError('Subclasses must implement this')

    def get_data_from_row(self, data: list[str]) -> list[any]:
        return [
            from_string(v, type_str)
            for v, type_str
            in zip(data, self.data_type_strs())
        ]

    def get_data_as_row(self) -> list[str]:
        """
        Returns `self.get_data()` as string list that can easily be added as row to a CSV file.
        """
        return self.as_row(self.get_data())

    def as_row(self, data: list, data_schema: list[str] = None) -> list[str]:
        """
        Returns the given `data` as string list that can easily be added as row to a CSV file.
        """
        if data_schema is None:
            data_schema = self.data_schema()
        return [
            to_string(v, type_str)
            for v, type_str
            in zip(data, self.data_type_strs(data_schema))
        ]

    def has_data(self) -> bool:
        return self.data is not None

    def get_data(self) -> list[any]:
        """
        Returns the last data collected by `self.update()`.
        """
        if self.has_data():
            return self.data
        else:
            raise ValueError(f'Data has not yet been read. {self.__str__()}')

    def set_data(self, data: list[any] | None):
        self.data = data

    def data_type_strs(self, data_schema: list[str] = None) -> list[str]:
        if data_schema is None:
            data_schema = self.data_schema()
        return [x.split('#', maxsplit=1)[0] for x in data_schema]

    @abstractmethod
    def get_log_file(self) -> Path:
        raise ValueError('Subclasses must implement this')

    @classmethod
    def get_log_dir(cls) -> Path:
        return logging_dir()

    def __str__(self) -> str:
        key_value_strings = [f'classname: {type(self).__name__}']
        for key, value in vars(self).items():
            key_value_strings.append(f'{key}: {value}')
        return ', '.join(key_value_strings)
